package org.embulk.input.filename;

import java.util.List;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.Files;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.FileVisitResult;
import java.nio.file.attribute.BasicFileAttributes;
import com.google.common.collect.ImmutableList;
import org.slf4j.Logger;
import org.embulk.config.TaskReport;
import org.embulk.config.Task;
import org.embulk.config.TaskSource;
import org.embulk.spi.Exec;
import org.embulk.spi.FileInputPlugin;
import org.embulk.spi.BufferAllocator;
import org.embulk.spi.TransactionalFileInput;
import org.embulk.spi.util.InputStreamTransactionalFileInput;
import org.embulk.standards.LocalFileInputPlugin;



class FilenameFileInputStream extends FileInputStream {
    static int MAX_NAME_LENGTH = 255;
    int n;
    byte[] name;

    FilenameFileInputStream(File file) throws FileNotFoundException {
        super(file);
        n = 0;
        name = file.getName().getBytes();
    }

    FilenameFileInputStream(String path) throws FileNotFoundException {
        super(path);
        n = 0;
        name = path.getBytes();
    }

    @Override
    public int read() throws IOException {
        if (n < name.length) {
            byte b = name[n];
            n++;
            return b;
        } else if (n < MAX_NAME_LENGTH) {
            n++;
            return 0;
        } else {
            return super.read();
        }
    }

    @Override
    public int read(byte[] b) throws IOException {
        return read(b, 0, b.length);
    }  

    @Override
    public int read(byte[] b, int off, int len) throws IOException {
        if (n < MAX_NAME_LENGTH) {
            int i = 0;
            int c; 
            for (; i < len; i++) {
                c = read();
                if (c == -1) {
                    if ( i == 0 ) {
                        return -1;
                    }
                    break;
                }
                b[off + i] = (byte)c;
            }
            return i;
         } else {
            return super.read(b, off, len);
         }
    }  
}

public class FilenameFileInputPlugin
        extends LocalFileInputPlugin
        implements FileInputPlugin
{

    private final Logger log = Exec.getLogger(getClass());

    private final static Path CURRENT_DIR = Paths.get(".").normalize();

    @Override
    public List<String> listFiles(PluginTask task)
    {
        Path pathPrefix = Paths.get(task.getPathPrefix()).normalize();
        final Path directory;
        final String fileNamePrefix;
        if (Files.isDirectory(pathPrefix)) {
            directory = pathPrefix;
            fileNamePrefix = "";
        } else {
            fileNamePrefix = pathPrefix.getFileName().toString();
            Path d = pathPrefix.getParent();
            directory = (d == null ? CURRENT_DIR : d);
        }

        final ImmutableList.Builder<String> builder = ImmutableList.builder();
        final String lastPath = task.getLastPath().orNull();
        try {
            log.info("Listing local files at directory '{}' filtering filename by prefix '{}'", directory.equals(CURRENT_DIR) ? "." : directory.toString(), fileNamePrefix);
            Files.walkFileTree(directory, new SimpleFileVisitor<Path>() {
                @Override
                public FileVisitResult preVisitDirectory(Path path, BasicFileAttributes attrs)
                {
                    if (path.equals(directory)) {
                        return FileVisitResult.CONTINUE;
                    } else if (lastPath != null && path.toString().compareTo(lastPath.substring(0, path.toString().length())) < 0) {
                        return FileVisitResult.SKIP_SUBTREE;
                    } else if (path.getFileName().toString().startsWith(".")) {
                        return FileVisitResult.SKIP_SUBTREE;
                    } else {
                        if (path.getFileName().toString().startsWith(fileNamePrefix)) {
                            return FileVisitResult.CONTINUE;
                        } else {
                            return FileVisitResult.SKIP_SUBTREE;
                        }
                    }
                }


                @Override
                public FileVisitResult visitFile(Path path, BasicFileAttributes attrs)
                {
                    if (lastPath != null && path.toString().compareTo(lastPath) <= 0) {
                        return FileVisitResult.CONTINUE;
                    } else if (path.getFileName().toString().startsWith(".")) {
                         return FileVisitResult.CONTINUE;
                    } else {
                        if (path.getFileName().toString().startsWith(fileNamePrefix)) {
                            builder.add(path.toString());
                        }
                        return FileVisitResult.CONTINUE;
                    }
               }
            });
        } catch (IOException ex) {
            throw new RuntimeException(String.format("Failed get a list of local files at '%s'", directory), ex);
        }
        return builder.build();
    }

    @Override
    public TransactionalFileInput open(TaskSource taskSource, int taskIndex)
    {
        final PluginTask task = taskSource.loadTask(PluginTask.class);
        final String path = task.getFiles().get(taskIndex);

        return new InputStreamTransactionalFileInput(
                task.getBufferAllocator(),
                new InputStreamTransactionalFileInput.Opener() {
                    public InputStream open() throws IOException
                    {
                        return new FilenameFileInputStream(path);
                    }
                })
        {
            @Override
            public void abort()
            { }

            @Override
            public TaskReport commit()
            {
                return Exec.newTaskReport();
            }
        };
    }
}
